//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "TScanResultLister.h"
#include "THasher.h"
#include "ScanUtils.h"
#include "SystemUtils.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
// class TScanResultLister : public TObject
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
__fastcall TScanResultLister::TRange::TRange(int AFirst, int ALast) :
	First(AFirst),
	Last(ALast)
{
	// Do nothing
}
//---------------------------------------------------------------------------
void __fastcall TScanResultLister::Initialize(TScanRecordList* ATargets,
	TScanOptions AOptions)
{
	try {
		FTargets = new TScanRecordList();
		Targets->AddExtractions(ATargets);
		FDuplicate = new TScanRecordList();
		Duplicate->Capacity = Targets->Count;
		FUnique = new TScanRecordList();
		Unique->Capacity = Targets->Count;
		FError = new TScanRecordList();
		Error->Capacity = Targets->Count;
		FRanges = new TTypedList<TRange>();
		Ranges->Add(new TRange(0, Targets->Count - 1));
		FComparers = new TTypedList<TScanComparer>();
		{
			Comparers->Add(new TScanSizeComparer());
		}
		if((AOptions.Properties & TScanOptions::PROPERTY_BASE_NAME) != 0) {
			Comparers->Add(new TScanBaseNameComparer());
		}
		if((AOptions.Properties & TScanOptions::PROPERTY_TYPE) != 0) {
			Comparers->Add(new TScanTypeComparer());
		}
		if((AOptions.Properties &
			TScanOptions::PROPERTY_MODIFICATION_TIME) != 0) {
			Comparers->Add(new TScanModificationTimeComparer());
		}
		{
			TScanHashComparer* scanHashComparer = NULL;
			try {
				scanHashComparer = new TScanHashComparer();
				scanHashComparer->OnHash = ScanHashComparerHash;
				Comparers->Add(scanHashComparer);
			}
			catch(...) {
				delete scanHashComparer;
				throw;
			}
		}
		FHasher = MakeHasher(AOptions.HashAlgorithm);
		FHashBuffer = new byte[HASH_BUFFER_SIZE];
	}
	catch(...) {
		Uninitialize();
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TScanResultLister::Uninitialize()
{
	delete[] FHashBuffer;
	delete FHasher;
	delete FComparers;
	delete FRanges;
	delete FError;
	delete FUnique;
	delete FDuplicate;
	delete FTargets;
}
//---------------------------------------------------------------------------
bool __fastcall TScanResultLister::ListNext()
{
	bool succeeded = false;
	for(;;) {
		// Get a range
		if(Ranges->Count == 0) {
			break;
		}
		TRange range = *Ranges->Last();
		Ranges->Delete(Ranges->Count - 1);
		if(range.First > range.Last) {
			continue;
		}

		// Find duplicates in the range
		int pivotIndex = FindPivot(range);
		if(pivotIndex < 0) {
			AddResults(range, Unique);
			continue;
		}
		Targets->Exchange(pivotIndex, range.First);
		pivotIndex = range.First;
		int equalCount = 1;
		int lessCount = 0;
		int greaterCount = 0;
		while(equalCount + lessCount + greaterCount <
				range.Last - range.First + 1) {
			int matchIndex = range.First + equalCount + lessCount;
			int difference = CompareScanRecord(
				Targets->Items[matchIndex],
				Targets->Items[pivotIndex],
				-1);
			if(difference < 0) {
				lessCount++;
			}
			else if(difference == 0) {
				Targets->Exchange(matchIndex, range.First + equalCount);
				equalCount++;
			}
			else if(difference > 0) {
				Targets->Exchange(matchIndex, range.Last - greaterCount);
				greaterCount++;
			}
			DoContinue();
		}
		AddResults(TRange(range.First, range.First + equalCount - 1),
			equalCount > 1 ? Duplicate : Unique);

		// Add new ranges
		Ranges->Add(new TRange(
			range.First + equalCount,
			range.First + equalCount + lessCount - 1));
		Ranges->Add(new TRange(
			range.First + equalCount + lessCount,
			range.First + equalCount + lessCount + greaterCount - 1));

		succeeded = true;
		break;
	}
	return succeeded;
}
//---------------------------------------------------------------------------
int __fastcall TScanResultLister::FindPivot(TRange ARange)
{
	int pivotIndex = -1;
	for(int i = 0; i < ARange.Last - ARange.First + 1; i++) {
		int matchIndex = ARange.First + i;
		if(!Targets->Items[matchIndex]->HasError() &&
				Targets->Items[matchIndex]->Primary) {
			pivotIndex = matchIndex;
			break;
		}
	}
	return pivotIndex;
}
//---------------------------------------------------------------------------
void __fastcall TScanResultLister::AddResults(TRange ARange,
	TScanRecordList* AResults)
{
	int group = AResults == Duplicate ?
		(Duplicate->Count == 0 ? 0 : (Duplicate->Last()->Group + 1)) :
		-1;
	for(int i = 0; i < ARange.Last - ARange.First + 1; i++) {
		TScanRecord* record = Targets->ReplaceAt(ARange.First + i, NULL);
		try {
			record->Group = record->HasError() ? -1 : group;
			(record->HasError() ? Error : AResults)->Add(record);
		}
		catch(...) {
			delete record;
			throw;
		}
	}
}
//---------------------------------------------------------------------------
int __fastcall TScanResultLister::CompareScanRecord(
	TScanRecord* AScanRecord1, TScanRecord* AScanRecord2, int ADefault)
{
	int difference = ADefault;
	for(int i = 0; i < Comparers->Count; i++) {
		difference = Comparers->Items[i]->CompareRecord(AScanRecord1,
			AScanRecord2, ADefault);
		if(difference != 0) {
			break;
		}
	}
	return difference;
}
//---------------------------------------------------------------------------
void __fastcall TScanResultLister::ScanHashComparerHash(TObject* ASender,
	AnsiString AFileName, AnsiString& AHash, bool& ASucceeded)
{
	ASucceeded = false;
	try {
		TFileStream* fileStream = NULL;
		try {
			fileStream = new TFileStream(AFileName,
				fmOpenRead | fmShareDenyWrite);
			fileStream->Position = 0;
			Hasher->Start();
			for(;;) {
				int dataLength = fileStream->Read(HashBuffer,
					HASH_BUFFER_SIZE);
				if(dataLength <= 0) {
					break;
				}
				Hasher->Update(HashBuffer, dataLength);
				DoContinue();
			}
			AHash = Hasher->Finish().AsText;
			ASucceeded = true;
		}
		__finally {
			delete fileStream;
		}
	}
	catch(EStreamError& e) {
		// Do nothing
	}
}
//---------------------------------------------------------------------------
THasher* __fastcall TScanResultLister::MakeHasher(int AHashAlgorithm) const
{
	THasher* hasher = NULL;
	switch(AHashAlgorithm) {
	case TScanOptions::HASH_ALGORITHM_MD5:
		hasher = new TMd5Hasher();
		break;
	case TScanOptions::HASH_ALGORITHM_SHA_1:
		hasher = new TSha1Hasher();
		break;
	case TScanOptions::HASH_ALGORITHM_SHA_256:
		hasher = new TSha256Hasher();
		break;
	default:
		throw EScanFailed("Invalid hash algorithm");
	}
	return hasher;
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TScanResultLister::TScanResultLister(TScanRecordList* ATargets,
		TScanOptions AOptions) :
	TWorker(),
	FTargets(NULL),
	FDuplicate(NULL),
	FUnique(NULL),
	FError(NULL),
	FRanges(NULL),
	FComparers(NULL),
	FHasher(NULL),
	FHashBuffer(NULL)
{
	Initialize(ATargets, AOptions);
}
//---------------------------------------------------------------------------
__fastcall TScanResultLister::~TScanResultLister()
{
	Uninitialize();
}
//---------------------------------------------------------------------------
void __fastcall TScanResultLister::List(TScanRecordList* ADuplicate,
	TScanRecordList* AUnique, TScanRecordList* AError)
{
	while(ListNext()) {
		DoProgress(Duplicate->Count + Unique->Count + Error->Count,
			Targets->Count);
		DoContinue();
	}
	ADuplicate->AddExtractions(Duplicate);
	AUnique->AddExtractions(Unique);
	AError->AddExtractions(Error);
}
//---------------------------------------------------------------------------

